import asyncio
import math
from datetime import datetime
from pathlib import Path

from uns_kit import UnsConfig, UnsMqttClient, TopicBuilder


def load_config() -> UnsConfig:
    cfg_path = Path("config.json")
    return UnsConfig.load(cfg_path)


def simulate_sensor_value(step: int) -> float:
    base_value = 42.0
    fast_cycle = math.sin(step / 5.0) * 3.0
    slow_cycle = math.sin(step / 25.0) * 6.0
    ripple = math.sin(step / 2.0 + math.pi / 4.0) * 0.5
    value = base_value + fast_cycle + slow_cycle + ripple
    return round(value, 2)


def prompt_bool(prompt: str, default_yes: bool = True) -> bool:
    suffix = " (Y/n) " if default_yes else " (y/N) "
    answer = input(prompt + suffix).strip().lower()
    if not answer:
        return default_yes
    return answer in ("y", "yes")


def prompt_int(prompt: str, default: int, suffix: str = "") -> int:
    answer = input(f"{prompt} (default {default}{suffix}) ").strip()
    try:
        return int(answer) if answer else default
    except ValueError:
        return default


def prompt_topic(prompt: str, default: str) -> str:
    answer = input(f"{prompt} (default {default}) ").strip()
    return answer or default


def log_info(message: str) -> None:
    timestamp = datetime.now().isoformat(timespec="seconds")
    print(f"{timestamp} [INFO] {message}")


async def main() -> None:
    cfg = load_config()
    host = f"{cfg.host}:{cfg.port}" if cfg.port else cfg.host

    if not prompt_bool(f"Would you like to continue with load-test on {host}?"):
        print("Load test aborted.")
        return

    count = prompt_int("How many iterations should be run?", 100, "")
    delay_ms = prompt_int("What should be the delay between intervals in milliseconds?", 0, " ms")
    topic = prompt_topic("Topic to publish to", "raw/data")
    retain = prompt_bool("Retain each published message so late subscribers see the last value?", True)
    delay_s = delay_ms / 1000.0

    log_info(f"Starting load test with {count} messages and {delay_ms} ms delay...")

    # Use a dedicated identity for load testing to avoid clashing with app processes.
    tb = TopicBuilder(package_name="uns-loadtest", package_version="0.0.0", process_name="load-tester")
    client = UnsMqttClient(
        cfg.host,
        port=cfg.port,
        username=cfg.username or None,
        password=cfg.password or None,
        tls=cfg.tls,
        client_id="uns-loadtest-client",
        topic_builder=tb,
        instance_name="py-load-test",
        reconnect_interval=1,
    )
    await client.connect()

    start = asyncio.get_event_loop().time()
    for i in range(count):
        now = asyncio.get_event_loop().time()
        sensor_value = simulate_sensor_value(i)
        raw_payload = f"{i},{int(now * 1000)},{sensor_value}"
        await client.publish_raw(topic, raw_payload, retain=retain)
        if delay_s:
            await asyncio.sleep(delay_s)

    log_info("Sleeping for 50ms.")
    await asyncio.sleep(0.05)
    end = asyncio.get_event_loop().time()
    duration = end - start
    rate = count / duration if duration > 0 else 0
    log_info(f"Load test completed in {duration:.2f} seconds.")
    log_info(f"Message rate: {rate:.2f} msg/s.")

    await client.close()


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import asyncio
import base64
import json
import os
import re
import shutil
import importlib.resources
from pathlib import Path
import subprocess
import urllib.error
import urllib.request
from typing import Any, Dict, Optional, Tuple

import click

from .client import UnsMqttClient
from .packet import UnsPacket
from .topic_builder import TopicBuilder


def common_options(func):
    func = click.option("--host", required=True, help="MQTT host (hostname or host:port)")(func)
    func = click.option("--port", type=int, help="MQTT port override")(func)
    func = click.option("--username", help="MQTT username")(func)
    func = click.option("--password", help="MQTT password")(func)
    func = click.option("--tls/--no-tls", default=False, show_default=True, help="Enable TLS")(func)
    func = click.option("--client-id", help="MQTT clientId")(func)
    func = click.option("--package-name", default="uns-kit", show_default=True, help="Package name for infra topics")(func)
    func = click.option("--package-version", default="0.0.1", show_default=True, help="Package version for infra topics")(func)
    func = click.option("--process-name", default="uns-process", show_default=True, help="Process name for infra topics")(func)
    func = click.option("--reconnect-interval", default=1.0, show_default=True, type=float, help="Reconnect backoff start (s)")(func)
    return func


@click.group(help="Lightweight UNS MQTT helper (Python).")
def cli():
    pass


@cli.command("publish", help="Publish a UNS data packet to a topic.")
@common_options
@click.option("--topic", required=True, help="MQTT topic (e.g. raw/data/)")
@click.option("--value", required=False, help="Value to send (stringified if not JSON).")
@click.option("--uom", default=None, help="Unit of measure.")
@click.option("--json", "json_value", default=None, help="JSON string to use as value.")
@click.option("--qos", type=int, default=0, show_default=True)
@click.option("--retain/--no-retain", default=False, show_default=True)
def publish_cmd(**opts):
    asyncio.run(_run_publish(**opts))


async def _run_publish(
    host: str,
    port: Optional[int],
    username: Optional[str],
    password: Optional[str],
    tls: bool,
    client_id: Optional[str],
    package_name: str,
    package_version: str,
    process_name: str,
    reconnect_interval: float,
    topic: str,
    value: Optional[str],
    uom: Optional[str],
    json_value: Optional[str],
    qos: int,
    retain: bool,
):
    tb = TopicBuilder(package_name, package_version, process_name)
    client = UnsMqttClient(
        host=host.split(":")[0],
        port=int(host.split(":")[1]) if ":" in host and port is None else port,
        username=username,
        password=password,
        tls=tls,
        client_id=client_id,
        topic_builder=tb,
        reconnect_interval=reconnect_interval,
    )
    await client.connect()

    payload_obj = json.loads(json_value) if json_value is not None else value
    packet = UnsPacket.data(value=payload_obj, uom=uom)
    await client.publish_packet(topic, packet, qos=qos, retain=retain)
    await client.close()


@cli.command("subscribe", help="Subscribe to one or more topics (resilient).")
@common_options
@click.option("--topic", "topic_filter", required=True, help="Topic filter (e.g. uns-infra/#)")
def subscribe_cmd(**opts):
    asyncio.run(_run_subscribe(**opts))


async def _run_subscribe(
    host: str,
    port: Optional[int],
    username: Optional[str],
    password: Optional[str],
    tls: bool,
    client_id: Optional[str],
    package_name: str,
    package_version: str,
    process_name: str,
    reconnect_interval: float,
    topic_filter: str,
):
    tb = TopicBuilder(package_name, package_version, process_name)
    client = UnsMqttClient(
        host=host.split(":")[0],
        port=int(host.split(":")[1]) if ":" in host and port is None else port,
        username=username,
        password=password,
        tls=tls,
        client_id=client_id,
        topic_builder=tb,
        reconnect_interval=reconnect_interval,
    )
    await client.connect()

    async for msg in client.resilient_messages(topic_filter):
        try:
            print(f"{msg.topic} {msg.payload.decode()}")
        except Exception:
            print(f"{msg.topic} <binary {len(msg.payload)} bytes>")


@cli.command("create", help="Create a new UNS Python app (default template).")
@click.argument("dest")
def create(dest: str):
    template_root = importlib.resources.files("uns_kit").joinpath("templates/default")
    dest_path = os.path.abspath(dest)
    project_name = Path(dest_path).name
    initial_app_version = "0.1.0"
    if os.path.exists(dest_path):
        raise click.ClickException(f"Destination already exists: {dest_path}")
    shutil.copytree(template_root, dest_path)
    pyproject_path = Path(dest_path) / "pyproject.toml"
    if pyproject_path.exists():
        try:
            initial_app_version = _read_poetry_version(pyproject_path)
        except Exception:
            pass
    # Personalize config.json with project identity (TS-style nested config).
    config_path = Path(dest_path) / "config.json"
    if config_path.exists():
        try:
            data = json.loads(config_path.read_text())
            infra = data.setdefault("infra", {})
            uns_cfg = data.setdefault("uns", {})

            sanitized = TopicBuilder.sanitize_topic_part(project_name)
            # Keep a human-readable packageName; the TopicBuilder will sanitize for MQTT.
            uns_cfg["packageName"] = project_name
            # Use sanitized processName to avoid invalid topic parts.
            uns_cfg["processName"] = sanitized
            # Set an initial packageVersion (controller/deploy tags may override later).
            uns_cfg["packageVersion"] = initial_app_version

            infra["clientId"] = sanitized
            config_path.write_text(json.dumps(data, indent=2))
        except Exception:
            pass
    # Personalize pyproject.toml so pip/poetry builds succeed with a project-specific name/module.
    if pyproject_path.exists():
        try:
            _personalize_pyproject(pyproject_path, project_name)
        except Exception:
            pass
    # Personalize package.json for PM2 metadata (optional; controller also writes this for python RTT nodes).
    pkg_path = Path(dest_path) / "package.json"
    if pkg_path.exists():
        try:
            pkg_data = json.loads(pkg_path.read_text())
            pkg_data["name"] = project_name
            pkg_data["version"] = initial_app_version
            pkg_path.write_text(json.dumps(pkg_data, indent=2))
        except Exception:
            pass
    click.echo(f"Created UNS Python app at {dest_path}")
    click.echo("Next steps:")
    click.echo(f"  1) cd {dest_path}")
    click.echo("  2) poetry install")
    click.echo("  3) poetry run python main.py")
    click.echo("  4) Edit config.json with your MQTT host/credentials")


@cli.command("write-config", help="Write a minimal config.json scaffold.")
@click.option("--path", default="config.json", show_default=True)
def write_config(path: str):
    _write_config_file(Path(path))
    click.echo(f"Wrote {path}")


@cli.command("configure-devops", help="Add Azure DevOps settings and pipeline to a project.")
@click.argument("dest", required=False, default=".")
@click.option("--overwrite/--no-overwrite", default=False, show_default=True, help="Overwrite existing azure-pipelines.yml")
def configure_devops(dest: str, overwrite: bool):
    dest_path = Path(dest).resolve()
    config_path = dest_path / "config.json"
    if not config_path.exists():
        click.echo(f"config.json not found at {config_path}, generating default.")
        _write_config_file(config_path)

    config = json.loads(config_path.read_text())
    org, project = _prompt_devops(config)
    config.setdefault("devops", {})
    config["devops"]["provider"] = "azure-devops"
    config["devops"]["organization"] = org
    config["devops"]["project"] = project
    config_path.write_text(json.dumps(config, indent=2))
    click.echo(f"Updated devops settings in {config_path}")

    pipeline_template = importlib.resources.files("uns_kit").joinpath("templates/azure-pipelines.yml")
    pipeline_target = dest_path / "azure-pipelines.yml"
    if pipeline_target.exists() and not overwrite:
        click.echo("azure-pipelines.yml already exists (skipped). Use --overwrite to replace.")
    else:
        shutil.copyfile(pipeline_template, pipeline_target)
        click.echo(f"Wrote {pipeline_target}")

    click.echo("DevOps configuration complete. Next steps:")
    click.echo("  1) Commit config.json and azure-pipelines.yml")
    click.echo("  2) Set AZURE_PAT in your CI or local env for pipeline access")


def _prompt_devops(config: dict) -> Tuple[str, str]:
    default_org = (config.get("devops") or {}).get("organization") or os.environ.get("AZURE_ORG") or ""
    default_proj = (config.get("devops") or {}).get("project") or os.environ.get("AZURE_PROJECT") or ""
    org = click.prompt("Azure DevOps organization", default=default_org or None, type=str)
    project = click.prompt("Azure DevOps project", default=default_proj or None, type=str)
    return org.strip(), project.strip()


def _write_config_file(path: Path) -> None:
    project_name = path.resolve().parent.name
    sanitized = TopicBuilder.sanitize_topic_part(project_name)
    data = {
        "infra": {
            "host": "localhost",
            "port": 1883,
            "username": "",
            "password": "",
            "tls": False,
            "clientId": sanitized,
            "mqttSubToTopics": [],
            "keepalive": 60,
            "clean": True,
        },
        "uns": {
            "packageName": project_name,
            "packageVersion": "0.1.0",
            "processName": sanitized,
        },
    }
    path.write_text(json.dumps(data, indent=2))


@cli.command("configure-vscode", help="Add VS Code settings for Python development.")
@click.argument("dest", required=False, default=".")
@click.option("--overwrite/--no-overwrite", default=False, show_default=True, help="Overwrite existing .vscode files")
def configure_vscode(dest: str, overwrite: bool):
    dest_path = Path(dest).resolve()
    vscode_dir = dest_path / ".vscode"
    vscode_dir.mkdir(parents=True, exist_ok=True)

    template_root = importlib.resources.files("uns_kit").joinpath("templates/vscode")
    for filename in ("settings.json", "launch.json", "extensions.json"):
        src = template_root / filename
        dst = vscode_dir / filename
        if dst.exists() and not overwrite:
            click.echo(f"{dst} already exists (skipped). Use --overwrite to replace.")
            continue
        shutil.copyfile(src, dst)
        click.echo(f"Wrote {dst}")

    click.echo("VS Code configuration complete.")

@cli.command("pull-request", help="Create an Azure DevOps pull request for a Python project (bumps version, commits, pushes, opens PR).")
@click.argument("dest", required=False, default=".")
def pull_request(dest: str):
    dest_path = Path(dest).resolve()

    # Require a git repo and clean tree (mirrors TS behavior).
    _assert_git_clean(dest_path)
    current_branch = _git_output(dest_path, ["rev-parse", "--abbrev-ref", "HEAD"]).strip()

    config = _load_json(dest_path / "config.json")
    devops = (config.get("devops") if isinstance(config, dict) else {}) or {}
    org = (devops.get("organization") or "").strip()
    project = (devops.get("project") or "").strip()
    if not org or not project:
        raise click.ClickException("Missing devops.organization/devops.project in config.json. Run `uns-kit-py configure-devops` first.")

    token = (os.environ.get("AZURE_PAT") or "").strip()
    if not token:
        token = click.prompt(
            f"Azure DevOps PAT (create one at https://dev.azure.com/{org}/_usersSettings/tokens)",
            hide_input=True,
            type=str,
        ).strip()
    if len(token) < 10:
        raise click.ClickException("AZURE_PAT token looks too short.")

    repo_name = _infer_repo_name(dest_path)
    repo = _azure_get_repo(org, project, repo_name, token)
    default_branch_ref = (repo.get("defaultBranch") or "refs/heads/master").strip()
    default_branch = default_branch_ref.replace("refs/heads/", "")
    if default_branch and current_branch == default_branch:
        raise click.ClickException(f"Refusing to create PR from default branch ({default_branch}). Create a feature branch first.")

    pyproject_path = dest_path / "pyproject.toml"
    current_version = _read_poetry_version(pyproject_path)
    new_version = click.prompt("Version tag for this PR", default=current_version, type=str).strip()
    if not new_version:
        raise click.ClickException("Version is required.")
    _assert_version_tag_unique(dest_path, new_version)

    _write_poetry_version(pyproject_path, new_version)
    _git(dest_path, ["add", "pyproject.toml"])
    _git(dest_path, ["commit", "-m", f"Set new production version: {new_version}"])
    _git(dest_path, ["push", "origin", current_branch])

    title = click.prompt("Title for the pull request", type=str).strip()
    description = click.prompt("Description for the pull request", default="", type=str)

    pr = _azure_create_pr(
        org=org,
        project=project,
        repo_id=str(repo.get("id") or ""),
        token=token,
        source_branch=current_branch,
        target_ref=default_branch_ref,
        title=title,
        description=description,
        label=new_version,
    )
    pr_id = pr.get("pullRequestId")
    pr_url = pr.get("url") or f"https://dev.azure.com/{org}/{project}/_git/{repo_name}/pullrequest/{pr_id}"
    click.echo(f"Pull request created: {pr_url}")


def _load_json(path: Path) -> Dict[str, Any]:
    if not path.exists():
        raise click.ClickException(f"Missing {path}.")
    try:
        data = json.loads(path.read_text())
        if not isinstance(data, dict):
            raise ValueError("not a JSON object")
        return data
    except Exception as exc:
        raise click.ClickException(f"Failed to read {path}: {exc}")


def _git(cwd: Path, args: list[str]) -> None:
    subprocess.run(["git", *args], cwd=str(cwd), check=True, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)


def _git_output(cwd: Path, args: list[str]) -> str:
    out = subprocess.check_output(["git", *args], cwd=str(cwd), stderr=subprocess.STDOUT)
    return out.decode("utf-8", errors="replace")


def _assert_git_clean(cwd: Path) -> None:
    status = _git_output(cwd, ["status", "--porcelain"]).strip()
    if status:
        raise click.ClickException("Repository needs to be clean. Please commit or stash changes before creating a PR.")


def _infer_repo_name(cwd: Path) -> str:
    # Prefer parsing origin remote; fallback to folder name.
    try:
        remote = _git_output(cwd, ["remote", "get-url", "origin"]).strip()
    except Exception:
        remote = ""

    if remote:
        # https://dev.azure.com/org/project/_git/repo
        m = re.search(r"/_git/([^/]+?)(?:\\.git)?$", remote)
        if m:
            return m.group(1)
        # git@ssh.dev.azure.com:v3/org/project/repo
        m = re.search(r":v3/[^/]+/[^/]+/([^/]+?)(?:\\.git)?$", remote)
        if m:
            return m.group(1)
        # fallback last path segment
        tail = remote.rstrip("/").split("/")[-1]
        if tail:
            return tail.replace(".git", "")

    return cwd.name


def _read_poetry_version(pyproject_path: Path) -> str:
    text = pyproject_path.read_text()
    in_section = False
    for line in text.splitlines():
        if line.strip() == "[tool.poetry]":
            in_section = True
            continue
        if in_section and line.startswith("[") and line.strip().endswith("]"):
            break
        if in_section:
            m = re.match(r'^version\\s*=\\s*\"([^\"]+)\"', line.strip())
            if m:
                return m.group(1)
    raise click.ClickException(f"Could not find tool.poetry version in {pyproject_path}")


def _write_poetry_version(pyproject_path: Path, new_version: str) -> None:
    lines = pyproject_path.read_text().splitlines()
    out: list[str] = []
    in_section = False
    changed = False
    for line in lines:
        if line.strip() == "[tool.poetry]":
            in_section = True
            out.append(line)
            continue
        if in_section and line.startswith("[") and line.strip().endswith("]"):
            in_section = False
        if in_section and not changed and re.match(r"^version\\s*=", line.strip()):
            out.append(f'version = "{new_version}"')
            changed = True
        else:
            out.append(line)
    if not changed:
        raise click.ClickException(f"Could not update version in {pyproject_path} (tool.poetry section missing?)")
    pyproject_path.write_text("\n".join(out) + "\n")


def _assert_version_tag_unique(cwd: Path, version: str) -> None:
    try:
        tags = _git_output(cwd, ["ls-remote", "--tags", "origin"])
    except Exception as exc:
        raise click.ClickException(f"Failed to list remote tags: {exc}")
    if f"refs/tags/{version}" in tags:
        raise click.ClickException(f"Version tag {version} already exists on the remote. Choose another version.")


def _azure_request(method: str, url: str, token: str, data: Optional[dict] = None) -> dict:
    raw = None if data is None else json.dumps(data).encode("utf-8")
    auth = base64.b64encode(f":{token}".encode("utf-8")).decode("ascii")
    req = urllib.request.Request(url, data=raw, method=method)
    req.add_header("Authorization", f"Basic {auth}")
    req.add_header("Accept", "application/json")
    if raw is not None:
        req.add_header("Content-Type", "application/json")
    try:
        with urllib.request.urlopen(req) as resp:
            body = resp.read().decode("utf-8", errors="replace")
            return json.loads(body) if body else {}
    except urllib.error.HTTPError as e:
        body = e.read().decode("utf-8", errors="replace") if hasattr(e, "read") else str(e)
        raise click.ClickException(f"Azure DevOps API error {e.code}: {body}")


def _azure_get_repo(org: str, project: str, repo_name: str, token: str) -> dict:
    url = f"https://dev.azure.com/{org}/{project}/_apis/git/repositories/{repo_name}?api-version=7.1-preview.1"
    return _azure_request("GET", url, token)


def _azure_create_pr(
    *,
    org: str,
    project: str,
    repo_id: str,
    token: str,
    source_branch: str,
    target_ref: str,
    title: str,
    description: str,
    label: str,
) -> dict:
    if not repo_id:
        raise click.ClickException("Azure repository id missing; cannot create PR.")
    url = f"https://dev.azure.com/{org}/{project}/_apis/git/repositories/{repo_id}/pullrequests?api-version=7.1-preview.1"
    payload = {
        "sourceRefName": f"refs/heads/{source_branch}",
        "targetRefName": target_ref,
        "title": title,
        "description": description,
        "labels": [{"name": label}],
    }
    return _azure_request("POST", url, token, payload)


def _to_distribution_name(name: str) -> str:
    # PyPI distribution name: keep it readable and normalized.
    value = re.sub(r"[^a-zA-Z0-9_-]+", "-", name).strip("-").lower()
    value = re.sub(r"-{2,}", "-", value)
    return value or "uns-py-app"


def _to_module_name(name: str) -> str:
    value = re.sub(r"[^a-zA-Z0-9_]+", "_", name).strip("_").lower()
    value = re.sub(r"_{2,}", "_", value)
    if not value:
        value = "uns_py_app"
    if value[0].isdigit():
        value = f"app_{value}"
    return value


def _personalize_pyproject(pyproject_path: Path, project_name: str) -> None:
    dist_name = _to_distribution_name(project_name)
    module_name = _to_module_name(project_name)

    text = pyproject_path.read_text()
    text = re.sub(r'(?m)^name\\s*=\\s*\"[^\"]+\"\\s*$', f'name = \"{dist_name}\"', text)
    text = re.sub(
        r'(?m)^packages\\s*=\\s*\\[\\s*\\{\\s*include\\s*=\\s*\"[^\"]+\"\\s*\\}\\s*\\]\\s*$',
        f'packages = [{{ include = \"{module_name}\" }}]',
        text,
    )
    pyproject_path.write_text(text)

    # Rename the placeholder package folder to match the configured include (if present).
    pkg_dir = pyproject_path.parent / "uns_py_app"
    target_dir = pyproject_path.parent / module_name
    if pkg_dir.exists() and module_name != "uns_py_app" and not target_dir.exists():
        pkg_dir.rename(target_dir)


def main():
    cli()


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
import tracemalloc
from datetime import datetime, timezone
from typing import Callable, Optional

from .client import UnsMqttClient
from .packet import UnsPacket
from .topic_builder import TopicBuilder


class StatusMonitor:
    """
    Periodically publishes process-level status topics:
    - active
    - heap-used
    - heap-total
    """

    def __init__(
        self,
        client: UnsMqttClient,
        topic_builder: TopicBuilder,
        active_supplier: Callable[[], bool],
        interval_s: float = 10.0,
    ) -> None:
        self._client = client
        self._topic_builder = topic_builder
        self._active_supplier = active_supplier
        self._interval_s = interval_s
        self._memory_task: Optional[asyncio.Task] = None
        self._status_task: Optional[asyncio.Task] = None
        self._running = False

    async def start(self) -> None:
        if self._running:
            return
        if not tracemalloc.is_tracing():
            tracemalloc.start()
        self._running = True
        self._memory_task = asyncio.create_task(self._publish_memory_loop())
        self._status_task = asyncio.create_task(self._publish_active_loop())

    async def stop(self) -> None:
        self._running = False
        for task in (self._memory_task, self._status_task):
            if task and not task.done():
                task.cancel()
                try:
                    await task
                except asyncio.CancelledError:
                    pass

    async def _publish_memory_loop(self) -> None:
        topic_base = self._topic_builder.process_status_topic
        heap_used_topic = f"{topic_base}heap-used"
        heap_total_topic = f"{topic_base}heap-total"
        while self._running:
            current, peak = tracemalloc.get_traced_memory()
            time = datetime.now(timezone.utc)
            heap_used = round(current / 1048576)
            heap_total = round(peak / 1048576)
            heap_used_packet = UnsPacket.data(
                value=heap_used,
                uom="MB",
                time=time,
            )
            heap_total_packet = UnsPacket.data(
                value=heap_total,
                uom="MB",
                time=time,
            )
            await self._client.publish_raw(heap_used_topic, UnsPacket.to_json(heap_used_packet))
            await self._client.publish_raw(heap_total_topic, UnsPacket.to_json(heap_total_packet))
            await asyncio.sleep(self._interval_s)

    async def _publish_active_loop(self) -> None:
        topic = self._topic_builder.active_topic()
        while self._running:
            time = datetime.now(timezone.utc)
            active_packet = UnsPacket.data(
                value=1 if self._active_supplier() else 0,
                uom="bit",
                time=time,
            )
            await self._client.publish_raw(topic, UnsPacket.to_json(active_packet))
            await asyncio.sleep(self._interval_s)



from __future__ import annotations

import re


class TopicBuilder:
    """
    Mirrors the TypeScript MqttTopicBuilder for infra topics.
    """

    def __init__(self, package_name: str, package_version: str, process_name: str):
        self.package_name = self.sanitize_topic_part(package_name)
        self.package_version = self.sanitize_topic_part(package_version)
        self.process_name = self.sanitize_topic_part(process_name)
        self._base = f"uns-infra/{self.package_name}/{self.package_version}/{self.process_name}/"
        if not re.match(r"^uns-infra/[^/]+/[^/]+/[^/]+/$", self._base):
            raise ValueError("processStatusTopic must follow 'uns-infra/<package>/<version>/<process>/'")

    @staticmethod
    def sanitize_topic_part(name: str) -> str:
        sanitized = re.sub(r"[^a-zA-Z0-9_-]+", "-", name)
        sanitized = re.sub(r"-{2,}", "-", sanitized)
        sanitized = sanitized.strip("-")
        return sanitized or "uns-process"

    @property
    def process_status_topic(self) -> str:
        return self._base

    def active_topic(self) -> str:
        return f"{self._base}active"

    def handover_topic(self) -> str:
        return f"{self._base}handover"

    def wildcard_active_topic(self) -> str:
        parts = self._base.strip("/").split("/")
        if len(parts) < 2:
            raise ValueError("processStatusTopic must follow 'uns-infra/<package>/<version>/<process>/'")
        return "/".join(parts[:2]) + "/+/+/active"

    def instance_status_topic(self, instance_name: str) -> str:
        sanitized = self.sanitize_topic_part(instance_name)
        return f"{self._base}{sanitized}/"

    @staticmethod
    def extract_base_topic(full_topic: str) -> str:
        parts = full_topic.split("/")
        if len(parts) < 4:
            raise ValueError("Invalid topic format. Expected 'uns-infra/<package>/<version>/<process>/'.")
        return "/".join(parts[:4]) + "/"



import asyncio
import json
from pathlib import Path

from uns_kit import UnsMqttClient, UnsPacket, TopicBuilder, UnsConfig


async def run():
    # Load TS-style nested config
    cfg = UnsConfig.load(Path("config.json"))
    tb = cfg.topic_builder()

    client = UnsMqttClient(
        cfg.host,
        port=cfg.port,
        username=cfg.username,
        password=cfg.password,
        tls=cfg.tls,
        client_id=cfg.client_id or f"{cfg.process_name}-py",
        topic_builder=tb,
        reconnect_interval=1,
    )
    await client.connect()

    # Publish a startup heartbeat to raw/data/ (can be changed)
    await client.publish_packet("raw/data/", UnsPacket.data(value="started", uom="state"))

    # Idle loop; adjust to your workload
    while True:
        await asyncio.sleep(5)


if __name__ == "__main__":
    asyncio.run(run())



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, Optional
import json


def isoformat(dt: datetime) -> str:
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    dt = dt.astimezone(timezone.utc)
    return dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"


@dataclass
class DataPayload:
    value: Any
    uom: Optional[str] = None
    time: Optional[str] = None
    dataGroup: Optional[str] = None
    intervalStart: Optional[str | int] = None
    intervalEnd: Optional[str | int] = None
    windowStart: Optional[str | int] = None
    windowEnd: Optional[str | int] = None
    eventId: Optional[str] = None
    deleted: Optional[bool] = None
    deletedAt: Optional[str | int] = None
    lastSeen: Optional[str | int] = None
    foreignEventKey: Optional[str] = None


@dataclass
class TablePayload:
    time: str
    columns: Any
    dataGroup: Optional[str] = None
    intervalStart: Optional[str | int] = None
    intervalEnd: Optional[str | int] = None
    windowStart: Optional[str | int] = None
    windowEnd: Optional[str | int] = None
    eventId: Optional[str] = None
    deleted: Optional[bool] = None
    deletedAt: Optional[str | int] = None
    lastSeen: Optional[str | int] = None


def _value_type(value: Any) -> str:
    if value is None:
        return "null"
    if isinstance(value, bool):
        return "boolean"
    if isinstance(value, (int, float)):
        return "number"
    if isinstance(value, str):
        return "string"
    return "object"


def _normalize_payload_fields(payload: Dict[str, Any], extra: Dict[str, Any]) -> None:
    mapping = {
        "data_group": "dataGroup",
        "interval_start": "intervalStart",
        "interval_end": "intervalEnd",
        "window_start": "windowStart",
        "window_end": "windowEnd",
        "event_id": "eventId",
        "deleted_at": "deletedAt",
        "last_seen": "lastSeen",
        "foreign_event_key": "foreignEventKey",
    }
    for key, value in extra.items():
        target = mapping.get(key, key)
        payload[target] = value


def _prune_none(payload: Dict[str, Any]) -> None:
    for key in list(payload.keys()):
        if payload[key] is None:
            payload.pop(key, None)


class UnsPacket:
    version: str = "1.3.0"

    @staticmethod
    def data(
        value: Any,
        uom: Optional[str] = None,
        time: Optional[datetime | str] = None,
        data_group: Optional[str] = None,
        created_at: Optional[datetime | str] = None,
        expires_at: Optional[datetime | str] = None,
        **extra: Any,
    ) -> Dict[str, Any]:
        resolved_time = (
            time if isinstance(time, str) else isoformat(time or datetime.now(timezone.utc))
        )
        payload = {
            "value": value,
            "uom": uom,
            "time": resolved_time,
            "dataGroup": data_group,
            "valueType": _value_type(value),
        }
        _normalize_payload_fields(payload, extra)
        _prune_none(payload)
        message: Dict[str, Any] = {"data": payload}
        if created_at is not None:
            message["createdAt"] = created_at if isinstance(created_at, str) else isoformat(created_at)
        if expires_at is not None:
            message["expiresAt"] = expires_at if isinstance(expires_at, str) else isoformat(expires_at)
        return {"version": UnsPacket.version, "message": message}

    @staticmethod
    def table(
        table: Optional[Dict[str, Any]] = None,
        *,
        columns: Optional[Any] = None,
        time: Optional[datetime | str] = None,
        data_group: Optional[str] = None,
        created_at: Optional[datetime | str] = None,
        expires_at: Optional[datetime | str] = None,
        **extra: Any,
    ) -> Dict[str, Any]:
        if table is None:
            if columns is None:
                raise ValueError("table() requires either table dict or columns list")
            resolved_time = (
                time if isinstance(time, str) else isoformat(time or datetime.now(timezone.utc))
            )
            payload = {
                "time": resolved_time,
                "columns": columns,
                "dataGroup": data_group,
            }
        else:
            payload = dict(table)
            if "time" not in payload:
                payload["time"] = (
                    time if isinstance(time, str) else isoformat(time or datetime.now(timezone.utc))
                )
            if data_group is not None and "dataGroup" not in payload:
                payload["dataGroup"] = data_group

        _normalize_payload_fields(payload, extra)
        _prune_none(payload)
        message: Dict[str, Any] = {"table": payload}
        if created_at is not None:
            message["createdAt"] = created_at if isinstance(created_at, str) else isoformat(created_at)
        if expires_at is not None:
            message["expiresAt"] = expires_at if isinstance(expires_at, str) else isoformat(expires_at)
        return {"version": UnsPacket.version, "message": message}

    @staticmethod
    def to_json(packet: Dict[str, Any]) -> str:
        return json.dumps(packet, separators=(",", ":"))

    @staticmethod
    def parse(packet_str: str) -> Optional[Dict[str, Any]]:
        try:
            return json.loads(packet_str)
        except Exception:
            return None

    @staticmethod
    def from_message(message: Dict[str, Any]) -> Dict[str, Any]:
        msg = dict(message)
        data = msg.get("data")
        if isinstance(data, dict):
            if "valueType" not in data and "value" in data:
                data = dict(data)
                data["valueType"] = _value_type(data.get("value"))
                msg["data"] = data
        return {"version": UnsPacket.version, "message": msg}



from __future__ import annotations

try:  # pragma: no cover - best-effort metadata fetch
    from importlib.metadata import version

    __version__ = version("uns-kit")
except Exception:  # fallback for editable/local
    __version__ = "0.0.0"




from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Dict, Optional

from .client import UnsMqttClient
from .packet import UnsPacket, isoformat
from .proxy import UnsProxy
from .topic_builder import TopicBuilder


class MessageMode(str, Enum):
    RAW = "raw"
    DELTA = "delta"
    BOTH = "both"


@dataclass
class LastValueEntry:
    value: Any
    uom: Optional[str]
    timestamp: datetime


class UnsMqttProxy(UnsProxy):
    def __init__(
        self,
        host: str,
        *,
        process_name: str,
        instance_name: str,
        package_name: str = "uns-kit",
        package_version: str = "0.0.1",
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        tls: bool = False,
        client_id: Optional[str] = None,
        keepalive: int = 60,
        clean_session: bool = True,
    ) -> None:
        self.topic_builder = TopicBuilder(package_name, package_version, process_name)
        self.instance_status_topic = self.topic_builder.instance_status_topic(instance_name)
        self.client = UnsMqttClient(
            host,
            port=port,
            username=username,
            password=password,
            tls=tls,
            client_id=client_id,
            keepalive=keepalive,
            clean_session=clean_session,
            topic_builder=self.topic_builder,
            instance_name=instance_name,
            publisher_active=True,
            subscriber_active=True,
        )
        super().__init__(self.client, self.instance_status_topic, instance_name)
        self._last_values: Dict[str, LastValueEntry] = {}
        self._sequence_ids: Dict[str, int] = {}

    async def connect(self) -> None:
        await self.client.connect()
        await self.start()

    async def close(self) -> None:
        await self.stop()
        await self.client.close()

    async def publish_message(self, topic: str, payload: str | bytes) -> None:
        await self.client.publish_raw(topic, payload)

    async def publish_packet(self, topic: str, packet: Dict[str, Any]) -> None:
        await self.client.publish_packet(topic, packet)

    async def publish_mqtt_message(self, mqtt_message: Dict[str, Any], mode: MessageMode = MessageMode.RAW) -> None:
        attrs = mqtt_message.get("attributes")
        if attrs is None:
            raise ValueError("mqtt_message must include attributes")
        if not isinstance(attrs, list):
            attrs = [attrs]

        base_topic = mqtt_message.get("topic", "")
        asset = mqtt_message.get("asset")
        asset_description = mqtt_message.get("assetDescription")
        object_type = mqtt_message.get("objectType")
        object_type_description = mqtt_message.get("objectTypeDescription")
        object_id = mqtt_message.get("objectId")

        for attr in attrs:
            attribute = attr.get("attribute")
            if attribute is None:
                raise ValueError("attribute is required")
            description = attr.get("description")
            tags = attr.get("tags")
            attribute_needs_persistence = attr.get("attributeNeedsPersistence")

            message = attr.get("message")
            if message is None:
                if "data" in attr:
                    message = {
                        "data": attr["data"],
                        **({"createdAt": attr["createdAt"]} if attr.get("createdAt") else {}),
                        **({"expiresAt": attr["expiresAt"]} if attr.get("expiresAt") else {}),
                    }
                elif "table" in attr:
                    message = {
                        "table": attr["table"],
                        **({"createdAt": attr["createdAt"]} if attr.get("createdAt") else {}),
                        **({"expiresAt": attr["expiresAt"]} if attr.get("expiresAt") else {}),
                    }
                else:
                    raise ValueError("Attribute entry must include exactly one of data/table/message")

            packet = UnsPacket.from_message(message)

            msg = {
                "topic": base_topic,
                "asset": asset,
                "assetDescription": asset_description,
                "objectType": object_type,
                "objectTypeDescription": object_type_description,
                "objectId": object_id,
                "attribute": attribute,
                "description": description or attribute,
                "tags": tags,
                "attributeNeedsPersistence": attribute_needs_persistence,
                "packet": packet,
            }

            if mode == MessageMode.RAW:
                await self._process_and_publish(msg, value_is_cumulative=False)
            elif mode == MessageMode.DELTA:
                delta_msg = dict(msg)
                delta_msg["attribute"] = f"{attribute}-delta"
                delta_msg["description"] = f"{msg['description']} (delta)"
                await self._process_and_publish(delta_msg, value_is_cumulative=True)
            elif mode == MessageMode.BOTH:
                await self._process_and_publish(msg, value_is_cumulative=False)
                delta_msg = dict(msg)
                delta_msg["attribute"] = f"{attribute}-delta"
                delta_msg["description"] = f"{msg['description']} (delta)"
                await self._process_and_publish(delta_msg, value_is_cumulative=True)

    def _resolve_object_identity(self, msg: Dict[str, Any]) -> None:
        topic = msg.get("topic", "")
        provided_type = msg.get("objectType")
        provided_id = msg.get("objectId")
        provided_asset = msg.get("asset")

        parts = [p for p in topic.split("/") if p]
        parsed_type = parts[-2] if len(parts) >= 2 else None
        parsed_id = parts[-1] if len(parts) >= 1 else None
        parsed_asset = parts[-3] if len(parts) >= 3 else None

        msg["objectType"] = provided_type or parsed_type
        msg["objectId"] = provided_id or parsed_id or "main"
        msg["asset"] = provided_asset or parsed_asset

    def _normalize_topic(self, topic: str) -> str:
        return topic if topic.endswith("/") else f"{topic}/"

    async def _process_and_publish(self, msg: Dict[str, Any], *, value_is_cumulative: bool) -> None:
        self._resolve_object_identity(msg)
        base_topic = self._normalize_topic(msg.get("topic", ""))

        packet = msg["packet"]
        message = packet.get("message", {})
        data = message.get("data")
        table = message.get("table")

        attribute_type = "Data" if data is not None else "Table" if table is not None else None
        data_group = ""
        if isinstance(data, dict):
            data_group = data.get("dataGroup") or ""
        if isinstance(table, dict):
            data_group = table.get("dataGroup") or ""

        await self.register_unique_topic(
            {
                "timestamp": isoformat(datetime.now(timezone.utc)),
                "topic": base_topic,
                "asset": msg.get("asset"),
                "assetDescription": msg.get("assetDescription"),
                "objectType": msg.get("objectType"),
                "objectTypeDescription": msg.get("objectTypeDescription"),
                "objectId": msg.get("objectId"),
                "attribute": msg.get("attribute"),
                "attributeType": attribute_type,
                "description": msg.get("description"),
                "tags": msg.get("tags"),
                "attributeNeedsPersistence": msg.get("attributeNeedsPersistence"),
                "dataGroup": data_group,
            }
        )

        publish_topic = (
            f"{base_topic}"
            f"{msg.get('asset') + '/' if msg.get('asset') else ''}"
            f"{msg.get('objectType') + '/' if msg.get('objectType') else ''}"
            f"{msg.get('objectId') + '/' if msg.get('objectId') else ''}"
            f"{msg.get('attribute')}"
        )

        seq_id = self._sequence_ids.get(base_topic, 0)
        self._sequence_ids[base_topic] = seq_id + 1
        packet["sequenceId"] = seq_id

        if isinstance(data, dict):
            time_value = data.get("time")
            if not time_value:
                time_value = UnsPacket.data(value=0)["message"]["data"]["time"]
                data["time"] = time_value
            current_time = datetime.fromisoformat(time_value.replace("Z", "+00:00"))
            new_value = data.get("value")
            new_uom = data.get("uom")
            last = self._last_values.get(publish_topic)
            if last:
                interval_ms = int((current_time - last.timestamp).total_seconds() * 1000)
                packet["interval"] = interval_ms
                if value_is_cumulative and isinstance(new_value, (int, float)) and isinstance(last.value, (int, float)):
                    delta = new_value - last.value
                    data["value"] = delta
                    data["time"] = isoformat(current_time)
                self._last_values[publish_topic] = LastValueEntry(new_value, new_uom, current_time)
                await self.client.publish_raw(publish_topic, UnsPacket.to_json(packet))
            else:
                self._last_values[publish_topic] = LastValueEntry(new_value, new_uom, current_time)
                # For delta mode with no previous value, skip to avoid bogus delta; otherwise publish.
                if not value_is_cumulative:
                    await self.client.publish_raw(publish_topic, UnsPacket.to_json(packet))
        elif isinstance(table, dict):
            await self.client.publish_raw(publish_topic, UnsPacket.to_json(packet))
        else:
            raise ValueError("packet.message must include data or table")



import asyncio
from datetime import datetime, timedelta, timezone
from pathlib import Path

from uns_kit import UnsConfig, UnsMqttClient, UnsProxyProcess
from uns_kit.packet import isoformat


async def main() -> None:
    cfg = UnsConfig.load(Path("config.json"))
    process = UnsProxyProcess(cfg.host, cfg)
    await process.start()
    print(f"[data-example] process client connected to {cfg.host}:{cfg.port or 1883}")

    out = await process.create_mqtt_proxy("py-output")
    out.client.publisher_active = True
    print(f"[data-example] output proxy connected to {cfg.host}:{cfg.port or 1883}")

    inp = UnsMqttClient(
        cfg.host,
        port=cfg.port,
        username=cfg.username or None,
        password=cfg.password or None,
        tls=cfg.tls,
        client_id=f"{cfg.process_name}-data-example-in",
        topic_builder=cfg.topic_builder(),
        instance_name="py-input",
        subscriber_active=True,
    )
    await inp.connect()
    print(f"[data-example] input client connected to {cfg.host}:{cfg.port or 1883}, subscribing to raw/#")

    topic = "enterprise/site/area/line/"
    asset = "asset"
    asset_description = "Sample asset"
    object_type = "energy-resource"
    object_id = "main"
    data_group = "sensor"

    try:
        async with inp.messages("raw/#") as messages:
            await out.client.publish_raw("raw/data", "", retain=True)
            print("[data-example] cleared retained raw/data; subscribed to 1 topic (raw/#); waiting for incoming raw/data...")

            async for msg in messages:
                payload = msg.payload.decode(errors="replace") if msg.payload else ""
                if str(msg.topic) != "raw/data":
                    continue

                parts = payload.split(",")
                if len(parts) < 3:
                    print(f"[data-example] skip malformed raw/data: {payload}")
                    continue
                try:
                    number_value = float(parts[0])
                    event_time_ms = int(parts[1])
                    sensor_value = float(parts[2])
                except ValueError:
                    print(f"[data-example] parse error, skip: {payload}")
                    continue

                event_time = datetime.fromtimestamp(event_time_ms / 1000, tz=timezone.utc)
                interval_start = isoformat(event_time - timedelta(seconds=1))
                interval_end = isoformat(event_time)
                time = isoformat(event_time)

                await out.publish_mqtt_message(
                    {
                        "topic": topic,
                        "asset": asset,
                        "assetDescription": asset_description,
                        "objectType": object_type,
                        "objectId": object_id,
                        "attributes": [
                            {
                                "attribute": "current",
                                "description": "Simulated current sensor value",
                                "data": {
                                    "dataGroup": data_group,
                                    "time": time,
                                    "intervalStart": interval_start,
                                    "intervalEnd": interval_end,
                                    "value": number_value,
                                    "uom": "A",
                                },
                            },
                            {
                                "attribute": "voltage",
                                "description": "Simulated voltage sensor value",
                                "data": {
                                    "dataGroup": data_group,
                                    "time": time,
                                    "intervalStart": interval_start,
                                    "intervalEnd": interval_end,
                                    "value": sensor_value,
                                    "uom": "V",
                                },
                            },
                        ],
                    }
                )
                print(f"[data-example] published transform at {time}")
    finally:
        await inp.close()
        await out.close()
        await process.stop()


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import asyncio
import uuid
from typing import List, Optional

from .client import UnsMqttClient
from .config import UnsConfig
from .status_monitor import StatusMonitor
from .topic_builder import TopicBuilder
from .uns_mqtt_proxy import UnsMqttProxy


class UnsProxyProcess:
    """
    Minimal Python equivalent of the TS UnsProxyProcess.
    Manages process-level status publishing and creates instance proxies.
    """

    def __init__(self, host: str, config: UnsConfig, process_name: Optional[str] = None, activate_delay_s: float = 10.0) -> None:
        self.config = config
        self.process_name = process_name or config.process_name
        self.process_id = uuid.uuid4().hex
        self.active = False
        self._activate_delay_s = activate_delay_s
        self.topic_builder = TopicBuilder(config.package_name, config.package_version, self.process_name)
        if config.client_id:
            process_client_id = f"{config.client_id}-{self.process_id}"
        else:
            process_client_id = f"{self.process_name}-{self.process_id}"
        self._client = UnsMqttClient(
            host,
            port=config.port,
            username=config.username or None,
            password=config.password or None,
            tls=config.tls,
            client_id=process_client_id,
            keepalive=config.keepalive,
            clean_session=config.clean_session,
            topic_builder=self.topic_builder,
            enable_status=False,
        )
        self._status_monitor = StatusMonitor(self._client, self.topic_builder, lambda: self.active)
        self._proxies: List[UnsMqttProxy] = []
        self._activate_task: Optional[asyncio.Task] = None

    async def start(self) -> None:
        await self._client.connect()
        await self._status_monitor.start()
        if self._activate_task is None or self._activate_task.done():
            self._activate_task = asyncio.create_task(self._activate_after_delay())

    async def stop(self) -> None:
        if self._activate_task and not self._activate_task.done():
            self._activate_task.cancel()
            try:
                await self._activate_task
            except asyncio.CancelledError:
                pass
        await self._status_monitor.stop()
        await self._client.close()

    def set_active(self, active: bool) -> None:
        self.active = active

    async def _activate_after_delay(self) -> None:
        await asyncio.sleep(self._activate_delay_s)
        if not self.active:
            self.active = True

    async def create_mqtt_proxy(
        self,
        instance_name: str,
        *,
        host: Optional[str] = None,
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        tls: Optional[bool] = None,
        client_id: Optional[str] = None,
    ) -> UnsMqttProxy:
        if client_id:
            resolved_client_id = client_id
        elif self.config.client_id:
            resolved_client_id = f"{self.config.client_id}-{instance_name}-{self.process_id}"
        else:
            resolved_client_id = f"{self.process_name}-{instance_name}-{self.process_id}"
        proxy = UnsMqttProxy(
            host or self.config.host,
            process_name=self.process_name,
            instance_name=instance_name,
            package_name=self.config.package_name,
            package_version=self.config.package_version,
            port=port if port is not None else self.config.port,
            username=username if username is not None else self.config.username,
            password=password if password is not None else self.config.password,
            tls=tls if tls is not None else self.config.tls,
            client_id=resolved_client_id,
        )
        await proxy.connect()
        self._proxies.append(proxy)
        return proxy



from __future__ import annotations

import asyncio
import json
from datetime import datetime, timezone
from typing import Any, Dict, Optional

from .client import UnsMqttClient
from .packet import isoformat


class UnsProxy:
    """
    Base proxy that tracks produced topics and periodically publishes the registry.
    """

    def __init__(self, client: UnsMqttClient, instance_status_topic: str, instance_name: str) -> None:
        self._client = client
        self._instance_status_topic = instance_status_topic
        self._instance_name = instance_name
        self._produced_topics: Dict[str, Dict[str, Any]] = {}
        self._publish_task: Optional[asyncio.Task] = None
        self._running = False

    async def start(self) -> None:
        if self._running:
            return
        self._running = True
        self._publish_task = asyncio.create_task(self._publish_loop())

    async def stop(self) -> None:
        self._running = False
        if self._publish_task and not self._publish_task.done():
            self._publish_task.cancel()
            try:
                await self._publish_task
            except asyncio.CancelledError:
                pass

    async def _publish_loop(self) -> None:
        while self._running:
            await self._emit_produced_topics()
            await asyncio.sleep(60)

    async def _emit_produced_topics(self) -> None:
        if not self._produced_topics:
            return
        payload = json.dumps(list(self._produced_topics.values()), separators=(",", ":"))
        await self._client.publish_raw(
            f"{self._instance_status_topic}topics",
            payload,
            retain=True,
        )

    async def register_unique_topic(self, topic_object: Dict[str, Any]) -> None:
        asset = topic_object.get("asset") or ""
        object_type = topic_object.get("objectType") or ""
        object_id = topic_object.get("objectId") or ""
        attribute = topic_object.get("attribute") or ""
        full_topic = f"{topic_object.get('topic', '')}{asset}/{object_type}/{object_id}/{attribute}"
        if full_topic not in self._produced_topics:
            topic_object.setdefault("timestamp", isoformat(datetime.now(timezone.utc)))
            self._produced_topics[full_topic] = topic_object
            await self._emit_produced_topics()



from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Tuple

from .topic_builder import TopicBuilder
from .version import __version__


@dataclass
class UnsConfig:
    # MQTT infra (matches TS nested config.infra)
    infra_host: str
    infra_port: Optional[int] = None
    infra_username: Optional[str] = None
    infra_password: Optional[str] = None
    infra_client_id: Optional[str] = None
    infra_tls: bool = False
    keepalive: int = 60
    clean_session: bool = True
    mqtt_sub_to_topics: Optional[List[str]] = None
    package_name: str = "uns-kit"
    package_version: str = __version__
    process_name: str = "uns-process"

    @staticmethod
    def load(path: Path) -> "UnsConfig":
        data = json.loads(path.read_text())

        if "infra" not in data or "uns" not in data:
            raise ValueError("config.json must include 'infra' and 'uns' sections (TS-style config)")

        infra = data.get("infra", {}) or {}
        uns_section = data.get("uns", {}) or {}
        pkg_name, pkg_version = _read_package_metadata(path)

        proto = (infra.get("protocol") or "").lower()
        tls = proto in ("mqtts", "ssl", "wss")

        host = infra.get("host") or (infra.get("hosts") or [None])[0]
        if not host:
            raise ValueError("infra.host (or hosts[0]) is required")

        return UnsConfig(
            infra_host=host,
            infra_port=infra.get("port"),
            infra_username=_none_if_empty(infra.get("username")),
            infra_password=_none_if_empty(infra.get("password")),
            infra_client_id=_none_if_empty(infra.get("clientId")),
            infra_tls=infra.get("tls") if infra.get("tls") is not None else tls,
            mqtt_sub_to_topics=infra.get("mqttSubToTopics"),
            keepalive=infra.get("keepalive", 60),
            clean_session=infra.get("clean", True),
            package_name=uns_section.get("packageName") or pkg_name or "uns-kit",
            package_version=uns_section.get("packageVersion") or pkg_version or __version__,
            process_name=uns_section.get("processName", "uns-process"),
        )

    def topic_builder(self) -> TopicBuilder:
        return TopicBuilder(self.package_name, self.package_version, self.process_name)

    # Convenience accessors for MQTT client creation
    @property
    def host(self) -> str:
        return self.infra_host

    @property
    def port(self) -> Optional[int]:
        return self.infra_port

    @property
    def username(self) -> Optional[str]:
        return self.infra_username

    @property
    def password(self) -> Optional[str]:
        return self.infra_password

    @property
    def client_id(self) -> Optional[str]:
        return self.infra_client_id

    @property
    def tls(self) -> bool:
        return self.infra_tls


def _none_if_empty(value: Optional[str]) -> Optional[str]:
    return None if value == "" else value


def _read_package_metadata(config_path: Path) -> Tuple[Optional[str], Optional[str]]:
    """
    Read name/version from the nearest package.json beside the config (if present).
    Falls back to None when not available or invalid.
    """
    pkg_path = config_path.parent / "package.json"
    if not pkg_path.exists():
        return None, None
    try:
        pkg = json.loads(pkg_path.read_text())
        name = pkg.get("name")
        version = pkg.get("version")
        return (name if isinstance(name, str) else None, version if isinstance(version, str) else None)
    except Exception:
        return None, None



from __future__ import annotations

import asyncio
import contextlib
from contextlib import asynccontextmanager
from datetime import datetime, timezone
import uuid
import socket
from typing import AsyncIterator, List, Optional

from asyncio_mqtt import Client, MqttError, Message, Will, Topic
from paho.mqtt.client import MQTTv5, MQTTv311

from .packet import UnsPacket
from .topic_builder import TopicBuilder


class UnsMqttClient:
    _exception_handler_installed = False
    _exception_handler_loop_id: int | None = None

    def __init__(
        self,
        host: str,
        *,
        topic_builder: TopicBuilder,
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        client_id: Optional[str] = None,
        tls: bool = False,
        keepalive: int = 60,
        clean_session: bool = True,
        reconnect_interval: float = 2.0,
        max_reconnect_interval: float = 30.0,
        instance_name: Optional[str] = None,
        publisher_active: Optional[bool] = None,
        subscriber_active: Optional[bool] = None,
        stats_interval: float = 60.0,
        enable_status: bool = True,
    ):
        self.host = host
        self.port = port
        self.username = username if username not in ("", None) else None
        self.password = password if password not in ("", None) else None
        if client_id:
            self.client_id = client_id
        else:
            base = topic_builder.process_name
            if instance_name:
                base = f"{base}-{instance_name}"
            self.client_id = f"{base}-{uuid.uuid4().hex[:8]}"
        self.tls = tls
        self.keepalive = keepalive
        self.clean_session = clean_session
        self.reconnect_interval = reconnect_interval
        self.max_reconnect_interval = max_reconnect_interval
        self.topic_builder = topic_builder
        self.instance_name = instance_name
        self.publisher_active = publisher_active
        self.subscriber_active = subscriber_active
        self.stats_interval = stats_interval
        self.enable_status = enable_status
        self._client: Optional[Client] = None
        self._status_task: Optional[asyncio.Task] = None
        self._stats_task: Optional[asyncio.Task] = None
        self._connected = asyncio.Event()
        self._closing = False
        self._connect_lock = asyncio.Lock()
        self._published_message_count = 0
        self._published_message_bytes = 0
        self._subscribed_message_count = 0
        self._subscribed_message_bytes = 0

        if self.instance_name:
            self.status_topic = self.topic_builder.instance_status_topic(self.instance_name)
        else:
            self.status_topic = self.topic_builder.process_status_topic

    def _install_exception_handler(self) -> None:
        loop = asyncio.get_running_loop()
        loop_id = id(loop)
        if UnsMqttClient._exception_handler_installed and UnsMqttClient._exception_handler_loop_id == loop_id:
            return
        original = loop.get_exception_handler()

        def handler(loop: asyncio.AbstractEventLoop, context: dict) -> None:
            exc = context.get("exception")
            message = context.get("message", "")
            if isinstance(exc, MqttError) and "Unexpected disconnect" in str(exc):
                # Swallow noisy disconnect futures; reconnect logic handles it.
                return
            if isinstance(exc, MqttError) and "Future exception was never retrieved" in message:
                return
            if original:
                original(loop, context)
            else:
                loop.default_exception_handler(context)

        loop.set_exception_handler(handler)
        UnsMqttClient._exception_handler_installed = True
        UnsMqttClient._exception_handler_loop_id = loop_id

    async def _connect_once(self) -> None:
        await self._probe_socket()
        will = Will(
            topic=f"{self.status_topic}alive",
            payload=b"",
            qos=0,
            retain=True,
        )
        base_kwargs = {
            "hostname": self.host,
            "port": self.port,
            "username": self.username,
            "password": self.password,
            "client_id": self.client_id,
            "keepalive": self.keepalive,
            "will": will,
        }
        if self.tls:
            try:
                import ssl
                base_kwargs["tls_context"] = ssl.create_default_context()
            except Exception:
                pass

        last_error: Exception | None = None
        # Prefer MQTT 3.1.1 for widest broker compatibility (some brokers
        # allow connect with MQTTv5 but then drop subscriptions unexpectedly).
        for protocol in (MQTTv311, MQTTv5):
            kwargs = dict(base_kwargs)
            kwargs["protocol"] = protocol
            if protocol == MQTTv5:
                kwargs["clean_start"] = self.clean_session
                kwargs.pop("clean_session", None)
            else:
                kwargs["clean_session"] = self.clean_session
                kwargs.pop("clean_start", None)
            try:
                try:
                    client = Client(**kwargs)
                except TypeError:
                    kwargs.pop("tls_context", None)
                    client = Client(**kwargs)
                await client.connect()
                self._client = client
                self._connected.set()
                return
            except Exception as exc:
                last_error = exc
                continue
        if last_error:
            raise last_error

    async def _probe_socket(self) -> None:
        try:
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(
                None,
                lambda: socket.create_connection((self.host, self.port or 1883), timeout=2),
            )
        except Exception as exc:
            raise MqttError(f"Cannot reach MQTT broker at {self.host}:{self.port or 1883}") from exc

    async def _ensure_connected(self) -> None:
        if self._connected.is_set():
            return
        async with self._connect_lock:
            if self._connected.is_set():
                return
            self._install_exception_handler()
            backoff = self.reconnect_interval
            while not self._closing:
                try:
                    await self._connect_once()
                    if self.enable_status:
                        if not self._status_task or self._status_task.done():
                            self._status_task = asyncio.create_task(self._publish_status_loop())
                            self._status_task.add_done_callback(self._handle_task_error)
                        if not self._stats_task or self._stats_task.done():
                            self._stats_task = asyncio.create_task(self._publish_stats_loop())
                            self._stats_task.add_done_callback(self._handle_task_error)
                    return
                except MqttError:
                    await asyncio.sleep(backoff)
                    backoff = min(backoff * 2, self.max_reconnect_interval)

    def _handle_task_error(self, task: asyncio.Task) -> None:
        try:
            task.exception()
        except asyncio.CancelledError:
            return
        except Exception:
            self._connected.clear()

    async def connect(self) -> None:
        await self._ensure_connected()

    async def close(self) -> None:
        self._closing = True
        if self._status_task:
            self._status_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._status_task
        if self._stats_task:
            self._stats_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._stats_task
        if self._client:
            with contextlib.suppress(Exception):
                await self._client.disconnect()
        self._connected.clear()

    async def publish_raw(self, topic: str, payload: str | bytes, *, qos: int = 0, retain: bool = False) -> None:
        await self._ensure_connected()
        for attempt in range(2):
            try:
                assert self._client
                payload_bytes = payload.encode() if isinstance(payload, str) else payload
                self._published_message_count += 1
                self._published_message_bytes += len(payload_bytes)
                await self._client.publish(topic, payload, qos=qos, retain=retain)
                return
            except MqttError:
                self._connected.clear()
                if attempt == 0:
                    await self._ensure_connected()
                else:
                    raise

    async def publish_packet(self, topic: str, packet: dict, *, qos: int = 0, retain: bool = False) -> None:
        payload = UnsPacket.to_json(packet)
        await self.publish_raw(topic, payload, qos=qos, retain=retain)

    @asynccontextmanager
    async def messages(self, topics: str | List[str]) -> AsyncIterator[AsyncIterator[Message]]:
        await self._ensure_connected()
        assert self._client
        async with self._client.messages() as messages:
            try:
                if isinstance(topics, str):
                    await self._client.subscribe(topics)
                else:
                    for t in topics:
                        await self._client.subscribe(t)
            except MqttError:
                self._connected.clear()
                raise

            async def wrapped() -> AsyncIterator[Message]:
                try:
                    async for msg in messages:
                        self._subscribed_message_count += 1
                        self._subscribed_message_bytes += len(msg.payload or b"")
                        yield msg
                except MqttError:
                    self._connected.clear()
                    return
                except asyncio.CancelledError:
                    return

            yield wrapped()

    async def resilient_messages(self, topics: str | List[str]) -> AsyncIterator[Message]:
        """
        Async generator that keeps the subscription alive across disconnects.
        """
        while not self._closing:
            await self._ensure_connected()
            try:
                async with self.messages(topics) as msgs:
                    async for msg in msgs:
                        yield msg
            except MqttError:
                self._connected.clear()
                await asyncio.sleep(self.reconnect_interval)
                continue
            except Exception:
                self._connected.clear()
                await asyncio.sleep(self.reconnect_interval)
                continue

    async def _publish_status_loop(self) -> None:
        uptime_topic = f"{self.status_topic}uptime"
        alive_topic = f"{self.status_topic}alive"
        publisher_topic = f"{self.status_topic}t-publisher-active"
        subscriber_topic = f"{self.status_topic}t-subscriber-active"
        start = asyncio.get_event_loop().time()
        try:
            while not self._closing:
                try:
                    now = asyncio.get_event_loop().time()
                    uptime_minutes = int((now - start) / 60)
                    time = datetime.now(timezone.utc)
                    alive_packet = UnsPacket.data(value=1, uom="bit", time=time)
                    uptime_packet = UnsPacket.data(value=uptime_minutes, uom="min", time=time)
                    await self.publish_raw(alive_topic, UnsPacket.to_json(alive_packet), qos=0, retain=False)
                    await self.publish_raw(uptime_topic, UnsPacket.to_json(uptime_packet), qos=0, retain=False)
                    if self.publisher_active is not None:
                        publisher_packet = UnsPacket.data(value=1 if self.publisher_active else 0, uom="bit", time=time)
                        await self.publish_raw(publisher_topic, UnsPacket.to_json(publisher_packet), qos=0, retain=False)
                    if self.subscriber_active is not None:
                        subscriber_packet = UnsPacket.data(value=1 if self.subscriber_active else 0, uom="bit", time=time)
                        await self.publish_raw(subscriber_topic, UnsPacket.to_json(subscriber_packet), qos=0, retain=False)
                except MqttError:
                    self._connected.clear()
                except Exception:
                    self._connected.clear()
                await asyncio.sleep(10)
        except asyncio.CancelledError:
            pass

    async def _publish_stats_loop(self) -> None:
        published_count_topic = f"{self.status_topic}published-message-count"
        published_bytes_topic = f"{self.status_topic}published-message-bytes"
        subscribed_count_topic = f"{self.status_topic}subscribed-message-count"
        subscribed_bytes_topic = f"{self.status_topic}subscribed-message-bytes"
        try:
            while not self._closing:
                try:
                    time = datetime.now(timezone.utc)
                    published_count_packet = UnsPacket.data(value=self._published_message_count, time=time)
                    published_bytes_packet = UnsPacket.data(
                        value=round(self._published_message_bytes / 1024),
                        uom="kB",
                        time=time,
                    )
                    subscribed_count_packet = UnsPacket.data(value=self._subscribed_message_count, time=time)
                    subscribed_bytes_packet = UnsPacket.data(
                        value=round(self._subscribed_message_bytes / 1024),
                        uom="kB",
                        time=time,
                    )
                    await self.publish_raw(published_count_topic, UnsPacket.to_json(published_count_packet), qos=0, retain=False)
                    await self.publish_raw(published_bytes_topic, UnsPacket.to_json(published_bytes_packet), qos=0, retain=False)
                    await self.publish_raw(subscribed_count_topic, UnsPacket.to_json(subscribed_count_packet), qos=0, retain=False)
                    await self.publish_raw(subscribed_bytes_topic, UnsPacket.to_json(subscribed_bytes_packet), qos=0, retain=False)
                    self._published_message_count = 0
                    self._published_message_bytes = 0
                    self._subscribed_message_count = 0
                    self._subscribed_message_bytes = 0
                except MqttError:
                    self._connected.clear()
                except Exception:
                    self._connected.clear()
                await asyncio.sleep(self.stats_interval)
        except asyncio.CancelledError:
            pass



from .topic_builder import TopicBuilder
from .packet import UnsPacket, DataPayload, TablePayload
from .client import UnsMqttClient
from .config import UnsConfig
from .status_monitor import StatusMonitor
from .uns_mqtt_proxy import UnsMqttProxy, MessageMode
from .proxy_process import UnsProxyProcess
from .version import __version__

__all__ = [
    "TopicBuilder",
    "UnsPacket",
    "DataPayload",
    "TablePayload",
    "UnsMqttClient",
    "UnsConfig",
    "StatusMonitor",
    "UnsMqttProxy",
    "MessageMode",
    "UnsProxyProcess",
    "__version__",
]
